#!/usr/bin/env node
/**
 * todo: createWriteStream owner/permissions
 */
/*global process, Buffer*/
(function() {
  "use strict";
  var fs = require('fs');
  var path = require('path');
  var url = require('url');

  var commander = require('commander');
  var ftpd = require('ftpd');
  var request = require('request');

  var createChunker = require('./lib/chunker');
  var LinkedHashMap = require('./lib/linked_hash_map');

  var CACHE_TTL = 800;
  var MAX_CACHE_LENGTH = 10;

  var getBaseName = path.basename;
  var getDirName = path.dirname;
  var parseUrl = url.parse;
  var slice = Array.prototype.slice;

  commander
    .version(require('./package.json').version)
    .option('-p, --port <port>', 'listen on port [2121]')
    .option('-b, --bind <address>', 'listen on address [127.0.0.1]')
    .parse(process.argv);

  var args = commander.args[0] === 'proxy' ? commander.args.slice(1) : commander.args.slice(0);
  var endpoint = args[0];
  var parsed = endpoint ? parseUrl(endpoint) : {};
  if (!parsed.protocol || !parsed.host || !parsed.pathname) {
    console.error('Error: Valid HTTP endpoint required');
    process.exit(1);
  }

  var config = {
    port: commander.port || 2121,
    bind: commander.bind || '127.0.0.1',
    silent: false
  };
  var ftp = createServer(config);
  ftp.listen(config.port, config.bind, function() {
    log('FTP server listening on ' + config.bind + ':' + config.port);
  });

  function createServer(config) {
    var ftp = new ftpd.FtpServer(config.bind, {
      pasvPortRangeStart: config.pasvPortRangeStart || 4000,
      pasvPortRangeEnd: config.pasvPortRangeEnd || 5000,
      getInitialCwd: function(connection, callback) {
        callback(null, '/');
      },
      getRoot: function(connection) {
        return '/';
      }
    });
    ftp.on('client:connected', function(connection) {
      config = Object.create(config);
      log('FTP connection from ' + connection.socket.remoteAddress);
      connection.on('command:user', function(username, success, failure) {
        config.username = username;
        success();
      });

      connection.on('command:pass', function(password, success, failure) {
        config.password = password;
        var proxy = new Proxy(config);
        proxy.readdir('/', function(error, list) {
          if (error) {
            log('Authentication failed:', error);
            return failure(error);
          } else {
            success(config.username, proxy);
          }
        });
      });

      connection.on('close', function() {
        log('Connection closed');
      });
    });
    return ftp;
  }


  function Proxy(config) {
    this.config = config;
    this.cache = new LinkedHashMap(MAX_CACHE_LENGTH);
  }

  Proxy.prototype = {
    cacheDirectoryList: function(data) {
      this.cache.addHead(data.path, {
        timestamp: Date.now(),
        data: data
      });
    },

    getCachedDirectoryList: function(path) {
      var cached = this.cache.get(path);
      var now = Date.now();
      if (cached && now - cached.timestamp < CACHE_TTL) {
        return cached.data;
      }
    },

    readdir: function(path, callback) {
      var cached = this.getCachedDirectoryList(path);
      if (cached) {
        process.nextTick(function() {
          callback(null, getNames(cached.children));
        });
        return;
      }
      this._get(toDirPath(path), function(error, data) {
        if (error) return callback(error);
        var children = data.children;
        this.cacheDirectoryList({path: path, children: children});
        callback(null, getNames(children));
      }.bind(this));
    },

    stat: function(path, callback) {
      var cached = this.getCachedDirectoryList(getDirName(path));
      if (cached) {
        var name = getBaseName(path);
        var children = cached.children;
        for (var i = 0, len = children.length; i < len; i++) {
          if (children[i].name === name) {
            callback(null, new StatsObject(children[i]));
            return;
          }
        }
      }
      this._get(path, function(error, data) {
        if (error) return callback(error);
        callback(null, new StatsObject(data));
      }.bind(this));
    },

    mkdir: function(path, mode, callback) {
      var dirname = getDirName(path);
      this._put(toDirPath(path), function(error, data) {
        if (error) return callback(error);
        this.cacheDirectoryList({path: dirname, children: data.children});
        callback();
      }.bind(this));
    },

    rmdir: function(path, callback) {
      var dirname = getDirName(path);
      this._delete(toDirPath(path), function(error, data) {
        if (error) return callback(error);
        this.cacheDirectoryList({path: dirname, children: data.children});
        callback();
      }.bind(this));
    },

    rename: function(src, dst, callback) {
      this._post(src, {path: dst}, function(error, data) {
        if (error) return callback(error);
        this.cacheDirectoryList({path: src, children: data.children});
        callback();
      }.bind(this));
    },

    unlink: function(path, callback) {
      this._delete(path, function(error, data) {
        if (error) return callback(error);
        this.cacheDirectoryList({path: getDirName(path), children: data.children});
        callback();
      }.bind(this));
    },

    createReadStream: function(path, opts) {
      var readStream = this._get(path, {dl: true}, null);
      readStream.on('response', function(response) {
        if (response.statusCode === 404) {
          var error = new Error('ENOENT, no such file');
          error.code = 'ENOENT';
          readStream.removeAllListeners('data');
          readStream.removeAllListeners('end');
          readStream.emit('error', error);
          response.connection.destroy();
        }
      });
      return readStream;
    },

    createWriteStream: function(path, opts) {
      var chunker = createChunker(4096);
      chunker.bytesWritten = 0;
      chunker.on('data', function(data) {
        var headers = {
          'Content-Type': 'application/octet-stream',
          'X-Append-File': (chunker.bytesWritten === 0) ? 'false' : 'true'
        };
        chunker.pause();
        this._put(path, headers, data, function(error, data) {
          if (error) {
            chunker.emit('error', error);
            return;
          }
          this.cacheDirectoryList({path: getDirName(path), children: data.children});
          chunker.bytesWritten += data.length;
          chunker.resume();
        }.bind(this));
      }.bind(this));
      process.nextTick(function() {
        chunker.emit('open');
      });
      return chunker;
    },

    _response: function(callback, error, response) {
      if (error) {
        callback(error);
      }
      var body = response.body.toString('utf8');
      var invalidatePath = response.headers['x-invalidate-cache'];
      if (invalidatePath) {
        this.cache.remove(invalidatePath);
      }
      var contentType = response.headers['content-type'] || '';
      if (contentType.split(';')[0] === 'application/json') {
        try {
          body = JSON.parse(body);
        } catch(e) {}
      }
      var statusCode = response.statusCode || 0;
      if (statusCode >= 400) {
        var bodyType = typeof body;
        if (bodyType === 'string' && body.match(/^[A-Z]+,/)) {
          error = new Error(body);
          error.code = body.split(',')[0];
        } else {
          var message = (bodyType === 'string') ? body : 'HTTP Error ' + statusCode;
          error = new Error(message);
          error.code = String(statusCode);
        }
        console.error('Received:', statusCode, error);
      }
      if (body === null || typeof body !== 'object') {
        body = {raw: body};
      }
      return callback(error, body)
    },

    _get: function(/*resource, headers, qs, callback*/) {
      var args = slice.call(arguments);
      var resource = args.shift();
      var callback = args.pop();
      var qs = args.pop();
      var headers = args.pop() || {};
      headers['X-Authorization'] = 'Basic ' + new Buffer(this.config.username + ':' + this.config.password).toString('base64');
      var opts = {
        method: 'GET',
        url: this._url(resource, qs),
        headers: headers,
        encoding: null
      };
      if (callback) {
        return request(opts, this._response.bind(this, callback));
      } else {
        return request(opts);
      }
    },

    _post: function(/*resource, headers, data, callback*/) {
      var args = slice.call(arguments);
      var resource = args.shift();
      var callback = args.pop();
      var data = args.pop() || {};
      var headers = args.pop() || {};
      headers['X-Authorization'] = 'Basic ' + new Buffer(this.config.username + ':' + this.config.password).toString('base64');
      if (!Buffer.isBuffer(data)) {
        data = new Buffer(JSON.stringify(data));
      }
      //workaround for error when sending empty request body
      if (data.length === 0) {
        headers['X-Empty-Body'] = 'true';
        data = new Buffer('~');
      }
      var opts = {
        method: 'POST',
        url: this._url(resource),
        headers: headers,
        body: data,
        encoding: null
      };
      if (callback) {
        return request(opts, this._response.bind(this, callback));
      } else {
        return request(opts);
      }
    },

    _put: function(/*resource, headers, data, callback*/) {
      var args = slice.call(arguments);
      var resource = args.shift();
      var callback = args.pop();
      var data = args.pop() || {};
      var headers = args.pop() || {};
      headers['X-Method-Override'] = 'PUT';
      headers['X-Authorization'] = 'Basic ' + new Buffer(this.config.username + ':' + this.config.password).toString('base64');
      if (!Buffer.isBuffer(data)) {
        data = new Buffer(JSON.stringify(data));
      }
      //workaround for error when sending empty request body
      if (data.length === 0) {
        headers['X-Empty-Body'] = 'true';
        data = new Buffer('~');
      }
      var opts = {
        method: 'POST',
        url: this._url(resource),
        headers: headers,
        body: data,
        encoding: null
      };
      if (callback) {
        return request(opts, this._response.bind(this, callback));
      } else {
        return request(opts);
      }
    },

    _delete: function(/*resource, headers, qs, callback*/) {
      var args = slice.call(arguments);
      var resource = args.shift();
      var callback = args.pop();
      var qs = args.pop();
      var headers = args.pop() || {};
      headers['X-Method-Override'] = 'DELETE';
      headers['X-Authorization'] = 'Basic ' + new Buffer(this.config.username + ':' + this.config.password).toString('base64');
      var opts = {
        method: 'GET',
        url: this._url(resource, qs),
        headers: headers,
        encoding: null
      };
      if (callback) {
        return request(opts, this._response.bind(this, callback));
      } else {
        return request(opts);
      }
    },

    _url: function(resource, qs) {
      var parts = ['path=' + encodeURIComponent(resource)];
      if (qs) {
        Object.keys(qs).forEach(function(key) {
          parts.push(encodeURIComponent(key) + '=' + encodeURIComponent(qs[key]));
        });
      }
      return endpoint + (endpoint.indexOf('?') === -1 ? '?' : '&') + parts.join('&');
    }
  };

  function StatsObject(item) {
    this.name = item.name || '';
    this.type = item.type;
    this.size = item.size || 0;
    this.ctime = new Date(Date.parse(item.dateCreated));
    this.mtime = new Date(Date.parse(item.dateLastModified));
    this.atime = new Date(Date.parse(item.dateLastAccessed));
  }
  StatsObject.prototype = {
    isFile: function() {
      return (this.type === 'file');
    },
    isDirectory: function() {
      return (this.type === 'directory');
    },
    isBlockDevice: function() { return false; },
    isCharacterDevice: function() { return false; },
    isSymbolicLink: function() { return false; },
    isFIFO: function() { return false; },
    isSocket: function() { return false; }
  };


  function getNames(items) {
    return items.map(function(item) {
      return item.name;
    });
  }

  function toDirPath(path) {
    return (path.slice(-1) === '/') ? path : path + '/'
  }

  function log() {
    if (!config.silent) {
      console.log.apply(console, arguments);
    }
  }
})();