/*global global, process, Buffer*/
(function() {
  var fs = process.binding('fs');
  var request = process.binding('request');
  var response = process.binding('response');

  var CHUNK_SIZE = 1024;
  var MAX_BODY_BUFFER = 8192;
  var OVERRIDE_ALLOWED = {
    'GET': {'HEAD': true, 'DELETE': true},
    'POST': {'PUT': true}
  };

  var USERS = {
    's92MYHBu1s': 's92MYHBu1s'
  };

  var headers = request.getHeaders();
  var url = request.getURL();
  //query should look something like `path=/path/to/file.jpg`
  var query = url.indexOf('?') >= 0 ? url.split('?')[1] : '';
  var queryParams = parseQueryString(query);

  var method = request.getMethod();
  var methodOverride = headers['x-method-override'];

  if (methodOverride in (OVERRIDE_ALLOWED[method] || {})) {
    method = methodOverride;
  }

  if (!queryParams.path) {
    if (method === 'GET') {
      sendResponse(200, 'OK', {'Content-Type': 'text/plain'}, 'SUCCESS. This endpoint is setup correctly.');
    }
    sendResponse(404, 'Not Found', {'Content-Type': 'text/plain'}, 'ENOENT, no such file or directory');
  }

  var isAuthorized = false;
  var auth = (headers['x-authorization'] || headers['authorization'] || '').split(' ');
  if (auth.length === 2 || auth[0] === 'Basic') {
    auth = new Buffer(auth[1], 'base64').toString().split(':');
    if (auth[0] in USERS && USERS[auth[0]] === auth.slice(1).join(':')) {
      isAuthorized = true;
    }
  }
  if (!isAuthorized) {
    sendResponse(401, 'Not Authorized', {'Content-Type': 'text/plain'}, '401 Not Authorized');
  }

  var path = normalizePath(queryParams.path, true);

  if (method === 'GET') {
    sendResource(path);
  }
  if (method === 'POST') {
    var newPath = renameResource(path);
    sendResource(getDirName(newPath), {'X-Invalidate-Cache': path});
  }
  if (method === 'PUT') {
    if (isDirPath(path)) {
      createDir(path);
      sendResource(getDirName(path));
    } else {
      saveFile(path);
      sendResource(getDirName(path));
    }
  }
  if (method === 'DELETE') {
    if (isDirPath(path)) {
      removeDir(path);
    } else {
      deleteFile(path);
    }
    sendResource(getDirName(path));
  }
  sendResponse(400, 'Bad Request', {'Content-Type': 'text/plain'}, 'Invalid HTTP Method');


  function sendResource(path, headers) {
    headers = headers || {};
    var info = getInfo(path);
    if (!info) {
      sendResponse(404, 'Not Found', {'Content-Type': 'text/plain'}, 'ENOENT, no such file or directory');
    }
    var isExpectingDir = isDirPath(path);
    if (isExpectingDir && info.type !== 'directory') {
      sendResponse(404, 'Not Found', {'Content-Type': 'text/plain'}, 'ENOENT, no such directory');
    }
    var isExpectingDownload = !!queryParams.dl;
    if (!isExpectingDownload) {
      //don't expose name of root directory
      if (path === '/') {
        info.name = '';
      }
      if (info.type === 'directory') {
        var childList = fs.getDirContents(mapPath(path));
        var prefix = (path.slice(-1) === '/') ? path : path + '/';
        info.children = childList.map(function(item) {
          return getInfo(prefix + item);
        });
      }
      headers['Content-Type'] = 'application/json';
      sendResponse(200, 'OK', headers, info);
    }
    if (info.type === 'directory') {
      sendResponse(409, 'Conflict', {'Content-Type': 'text/plain'}, 'EISDIR, path is a directory');
    }
    var readStream = fs.createReadStream(mapPath(path));
    headers['Content-Type'] = 'application/octet-stream';
    headers['Length'] = info.size;
    response.writeHead(200, 'OK', headers);
    readStream.on('data', function(data) {
      response.write(data);
    });
    readStream.on('end', function() {
      response.end();
    });
    readStream.read();
  }

  function saveFile(path) {
    var info = getInfo(getDirName(path));
    if (!info || info.type !== 'directory') {
      sendResponse(404, 'Not Found', {'Content-Type': 'text/plain'}, 'ENOENT, no such directory');
    }
    info = getInfo(path);
    if (info && info.type !== 'file') {
      sendResponse(409, 'Conflict', {'Content-Type': 'text/plain'}, 'EISDIR, path is a directory');
    }
    //workaround for error when sending empty request body
    if (headers['x-empty-body'] === 'true') {
      var data = new Buffer(0);
    } else {
      data = getRawBody();
    }
    var isAppend = (headers['x-append-file'] === 'true');
    fs.writeFile(mapPath(path), data, {append: isAppend});
  }

  function renameResource(path) {
    var info = getInfo(path);
    if (!info) {
      sendResponse(404, 'Not Found', {'Content-Type': 'text/plain'}, 'ENOENT, no such file or directory');
    }
    var srcType = info.type;
    var body = parseJSONBody();
    var dstPath = body.path;
    if (!dstPath || dstPath.indexOf('/') !== 0) {
      sendResponse(400, 'Bad Request', {'Content-Type': 'text/plain'}, '400 Invalid or Missing Destination');
    }
    dstPath = normalizePath(dstPath);
    info = getInfo(getDirName(dstPath));
    if (!info || info.type !== 'directory') {
      //todo: 409 Conflict?
      sendResponse(404, 'Not Found', {'Content-Type': 'text/plain'}, 'ENOENT, no such destination directory');
    }
    info = getInfo(dstPath);
    if (srcType === 'file') {
      var overwriteDestination = headers['x-overwrite'] === 'true';
      if (info && info.type === 'file' && !overwriteDestination) {
        sendResponse(409, 'Conflict', {'Content-Type': 'text/plain'}, 'EEXIST, destination already exists');
      }
      if (info && info.type !== 'file') {
        sendResponse(409, 'Conflict', {'Content-Type': 'text/plain'}, 'EISDIR, destination is a directory');
      }
      fs.moveFile(mapPath(path), mapPath(dstPath));
    } else {
      if (info) {
        sendResponse(409, 'Conflict', {'Content-Type': 'text/plain'}, 'EEXIST, destination already exists');
      }
      fs.moveDir(mapPath(path), mapPath(dstPath));
    }
    return dstPath;
  }

  function createDir(path) {
    var info = getInfo(getDirName(path));
    if (!info || info.type !== 'directory') {
      sendResponse(404, 'Not Found', {'Content-Type': 'text/plain'}, 'ENOENT, no such directory');
    }
    if (getInfo(path)) {
      sendResponse(409, '409 Conflict', {'Content-Type': 'text/plain'}, 'EEXIST, file or directory already exists');
    }
    fs.createDir(mapPath(path));
  }

  function removeDir(path) {
    var info = getInfo(path);
    if (!info || info.type !== 'directory') {
      sendResponse(404, 'Not Found', {'Content-Type': 'text/plain'}, 'ENOENT, no such directory');
    }
    var list = fs.getDirContents(mapPath(path));
    if (list.length) {
      sendResponse(409, '409 Conflict', {'Content-Type': 'text/plain'}, 'ENOTEMPTY, directory not empty');
    }
    fs.removeDir(mapPath(path));
  }

  function deleteFile(path) {
    var info = getInfo(path);
    if (!info || info.type !== 'file') {
      sendResponse(404, 'Not Found', {'Content-Type': 'text/plain'}, 'ENOENT, no such file');
    }
    fs.deleteFile(mapPath(path));
  }


  function sendResponse(statusCode, statusText, headers, body) {
    // just in case we're behind cloudflare or some proxy
    headers['Cache-Control'] = 'private, no-cache, no-transform';
    if (typeof body === 'string') {
      body = body + '\n';
    } else {
      body = JSON.stringify(body, null, 2) + '\n';
    }
    headers['Content-Length'] = body.length;
    response.writeHead(statusCode, statusText, headers);
    response.write(body);
    response.end();
  }

  function getInfo(path) {
    var info = null;
    try {
      info = fs.getInfo(mapPath(path));
      info.path = path;
    } catch(e) {
      if (e.code !== 'ENOENT') {
        throw e;
      }
    }
    return info;
  }

  function isDirPath(path) {
    return (path.slice(-1) === '/');
  }

  function getDirName(path) {
    path = path.slice(1);
    var trailingSlash = '';
    if (isDirPath(path)) {
      trailingSlash = '/';
      path = path.slice(-1);
    }
    var dirName = '/' + path.split('/').slice(0, -1).join('/');
    if (!isDirPath(dirName)) {
      dirName = dirName + trailingSlash;
    }
    return dirName;
  }

  function mapPath(path) {
    //path is expected to be already normalized to one of:
    //  `/directory/file.ext`
    //  `/directory/`
    //  `/`
    path = isDirPath(path) ? path.slice(0, -1) : path;
    return '.' + path;
  }

  function normalizePath(path, decode) {
    if (decode) {
      path = decodeURIComponent(path);
    }
    var suffix = isDirPath(path) ? '/' : '';
    while (path.charAt(0) === '/') {
      path = path.slice(1);
    }
    var parts = [];
    path.split('/').forEach(function(part) {
      if (part === '' || part === '.' || part === '..') return;
      parts.push(part);
    });
    path = parts.join('/') + suffix;
    return (path === '/') ? path : '/' + path;
  }

  function parseQueryString(query) {
    var parts = query.split('&');
    var result = {};
    parts.forEach(function(part) {
      var parts = part.split('=');
      var key = decodeURIComponent(parts[0]);
      result[key.toLowerCase()] = decodeURIComponent(parts[1] || '');
    });
    return result;
  }

  function parseJSONBody() {
    var rawBody = getRawBody('utf8');
    try {
      var body = JSON.parse(rawBody);
    } catch(e) {}
    if (!body || typeof body !== 'object') {
      sendResponse(400, 'Bad Request', {'Content-Type': 'text/plain'}, '400 Malformed JSON body');
    }
    return body;
  }

  function getRawBody(encoding) {
    var chunks = [];
    var bytesRead = 0;
    var chunk = request.read(CHUNK_SIZE);
    while (chunk !== null && (bytesRead += chunk.length) <= MAX_BODY_BUFFER) {
      chunks.push(chunk);
      chunk = request.read(CHUNK_SIZE);
    }
    if (bytesRead > MAX_BODY_BUFFER) {
      sendResponse(413, 'Request Entity Too Large', {'Content-Type': 'text/plain'}, '413 Request Entity Too Large');
    }
    var rawBody = Buffer.concat(chunks);
    if (encoding) {
      rawBody = rawBody.toString(encoding);
    }
    return rawBody;
  }

})();